/*
 * HelloQT2JNI.cpp
 *
 *  Created on: 14.06.2015
 *      Author: moozoom
 */


#include <jni.h>
#include <iostream>
#include "HelloQT2JNI.h"

#include "qtstaticlibrary.h"

using namespace std;

JNIEXPORT void JNICALL Java_HelloQT2JNI_sayHello__(JNIEnv *, jobject)
{
   cout << "Hello World!" << endl;

   QTStaticLibrary hello = QTStaticLibrary();
   hello.hello();

   return;
}

// Quelle: https://docs.oracle.com/javase/8/docs/technotes/guides/jni/spec/types.html
JNIEXPORT jstring JNICALL Java_HelloQT2JNI_sayHello__Ljava_lang_String_2(JNIEnv *env, jobject thisObj, jstring inJNIStr)
{
	// Step 1: Convert the JNI String (jstring) into C-String (char*)
	const char *inCStr = env->GetStringUTFChars(inJNIStr, NULL);
	if (NULL == inCStr) return NULL;

	// Step 2: Perform its intended operations
	cout << "In C++, the received string is: " << inCStr << endl;
	env->ReleaseStringUTFChars(inJNIStr, inCStr);  // release resources

	// Prompt user for a C++ string
	string outCppStr;
	cout << "Enter a String: ";
	getline(cin, outCppStr);
	//cin >> outCppStr;

	// Step 3: Convert the C++ string to C-string, then to JNI String (jstring) and return
	return env->NewStringUTF(outCppStr.c_str());
}

JNIEXPORT jdouble JNICALL Java_HelloQT2JNI_average(JNIEnv *, jobject, jint n1, jint n2)
{
	jdouble result;
	cout << "In C++, the numbers are " << n1 << " and " << n2 << endl;
	result = ((jdouble)n1 + n2) / 2.0;
	// jint is mapped to int, jdouble is mapped to double
	return result;
}

JNIEXPORT jdoubleArray JNICALL Java_HelloQT2JNI_sumAndAverage(JNIEnv *env, jobject, jintArray inJNIArray)
{
	// Step 1: Convert the incoming JNI jintarray to C's jint[]
	   jint *inCArray = env->GetIntArrayElements(inJNIArray, NULL);
	   if (NULL == inCArray) return NULL;
	   jsize length = env->GetArrayLength(inJNIArray);

	   // Step 2: Perform its intended operations
	   jint sum = 0;
	   int i;
	   for (i = 0; i < length; i++) {
	      sum += inCArray[i];
	   }
	   jdouble average = (jdouble)sum / length;
	   env->ReleaseIntArrayElements(inJNIArray, inCArray, 0); // release resources

	   jdouble outCArray[] = {sum, average};

	   // Step 3: Convert the C's Native jdouble[] to JNI jdoublearray, and return
	   jdoubleArray outJNIArray = env->NewDoubleArray(2);  // allocate
	   if (NULL == outJNIArray) return NULL;
	   env->SetDoubleArrayRegion(outJNIArray, 0 , 2, outCArray);  // copy
	   return outJNIArray;
}

JNIEXPORT void JNICALL Java_HelloQT2JNI_modifyInstanceVariable(JNIEnv *env, jobject thisObj)
{
	// Get a reference to this object's class
	jclass thisClass = env->GetObjectClass(thisObj);

	// int
	// Get the Field ID of the instance variables "intNumber"
	jfieldID fidNumber = env->GetFieldID(thisClass, "intNumber", "I");
	if (NULL == fidNumber) return;

	// Get the int given the Field ID
	jint number = env->GetIntField(thisObj, fidNumber);
	cout << "In C++, the int is " << number << endl;

	// Change the variable
	number = 99;
	env->SetIntField(thisObj, fidNumber, number);

	// Get the Field ID of the instance variables "message"
	jfieldID fidMessage = env->GetFieldID(thisClass, "message", "Ljava/lang/String;");
	if (NULL == fidMessage) return;

	// String
	// Get the object given the Field ID
	jstring message = (jstring)env->GetObjectField(thisObj, fidMessage);

	// Create a C-string with the JNI String
	const char *cStr = env->GetStringUTFChars(message, NULL);
	if (NULL == cStr) return;

	cout << "In C++, the string is " << cStr << endl;
	env->ReleaseStringUTFChars(message, cStr);

	// Create a new C-string and assign to the JNI string
	message = env->NewStringUTF("Hello from C++");
	if (NULL == message) return;

	// modify the instance variables
	env->SetObjectField(thisObj, fidMessage, message);

	return;
}

JNIEXPORT void JNICALL Java_HelloQT2JNI_modifyStaticVariable(JNIEnv *env, jobject thisObj)
{
	// Get a reference to this object's class
	jclass cls = env->GetObjectClass(thisObj);

	// Read the int static variable and modify its value
	jfieldID fidNumber = env->GetStaticFieldID(cls, "doubleNumber", "D");
	if (NULL == fidNumber) return;

	jdouble number = env->GetStaticDoubleField(cls, fidNumber);
	cout << "In C++, the double is " << number << endl;
	number = 77.88;
	env->SetStaticDoubleField(cls, fidNumber, number);

	return;
}

JNIEXPORT void JNICALL Java_HelloQT2JNI_nativeMethod(JNIEnv *env, jobject thisObj)
{
	// Get a class reference for this object
   jclass thisClass = env->GetObjectClass(thisObj);

   // Get the Method ID for method "callback", which takes no arg and return void
   jmethodID midCallBack = env->GetMethodID(thisClass, "callback", "()V");
   if (NULL == midCallBack) return;

   cout << "In C++, call back Java's callback()" << endl;
   // Call back the method (which returns void), baed on the Method ID
   env->CallVoidMethod(thisObj, midCallBack);

   jmethodID midCallBackStr = env->GetMethodID(thisClass, "callback", "(Ljava/lang/String;)V");
   if (NULL == midCallBackStr) return;

   cout << "In C++, call back Java's called(String)" << endl;
   jstring message = env->NewStringUTF("Hello from C++");
   env->CallVoidMethod(thisObj, midCallBackStr, message);

   jmethodID midCallBackAverage = env->GetMethodID(thisClass, "callbackAverage", "(II)D");
   if (NULL == midCallBackAverage) return;

   jdouble average = env->CallDoubleMethod(thisObj, midCallBackAverage, 2, 3);
   cout << "In C++, the average is " << average << endl;

   jmethodID midCallBackStatic = env->GetStaticMethodID(thisClass, "callbackStatic", "()Ljava/lang/String;");
   if (NULL == midCallBackStatic) return;

   jstring resultJNIStr = (jstring)env->CallStaticObjectMethod(thisClass, midCallBackStatic);
   const char *resultCStr = env->GetStringUTFChars(resultJNIStr, NULL);
   if (NULL == resultCStr) return;

   cout << "In C++, the returned string is " << resultCStr << endl;
   env->ReleaseStringUTFChars(resultJNIStr, resultCStr);

   return;
}

JNIEXPORT jobject JNICALL Java_HelloQT2JNI_getIntegerObject(JNIEnv *env, jobject thisObj, jint number)
{
	// Get a class reference for java.lang.Integer
	jclass cls = env->FindClass("java/lang/Integer");

	// Get the Method ID of the constructor which takes an int
	jmethodID midInit = env->GetMethodID(cls, "<init>", "(I)V");
	if (NULL == midInit) return NULL;

	// Call back constructor to allocate a new instance, with an int argument
	jobject newObj = env->NewObject(cls, midInit, number);

	// Try runnning the toString() on this newly create object
	jmethodID midToString = env->GetMethodID(cls, "toString", "()Ljava/lang/String;");
	if (NULL == midToString) return NULL;

	jstring resultStr = (jstring)env->CallObjectMethod(newObj, midToString);
    const char *resultCStr = env->GetStringUTFChars(resultStr, NULL);
    cout << "In C++, the number is " << resultCStr << endl;

    return newObj;
}

JNIEXPORT jobjectArray JNICALL Java_HelloQT2JNI_objSumAndAverage(JNIEnv *env, jobject thisObj, jobjectArray inJNIArray)
{
	// Get a class reference for java.lang.Integer
    jclass classInteger = env->FindClass("java/lang/Integer");
    // Use Integer.intValue() to retrieve the int
    jmethodID midIntValue = env->GetMethodID(classInteger, "intValue", "()I");
    if (NULL == midIntValue) return NULL;

    // Get the value of each Integer object in the array
    jsize length = env->GetArrayLength(inJNIArray);
    jint sum = 0;
    int i;
    for (i = 0; i < length; i++)
    {
	   jobject objInteger = env->GetObjectArrayElement(inJNIArray, i);
	   if (NULL == objInteger) return NULL;

	   jint value = env->CallIntMethod(objInteger, midIntValue);
	   sum += value;
    }
    double average = (double)sum / length;
    cout << "In C++, the sum is " << sum << endl;
    cout << "In C++, the average is " << average << endl;

    // Get a class reference for java.lang.Double
    jclass classDouble = env->FindClass("java/lang/Double");

    // Allocate a jobjectArray of 2 java.lang.Double
    jobjectArray outJNIArray = env->NewObjectArray(2, classDouble, NULL);

    // Construct 2 Double objects by calling the constructor
    jmethodID midDoubleInit = env->GetMethodID(classDouble, "<init>", "(D)V");
    if (NULL == midDoubleInit) return NULL;

    jobject objSum = env->NewObject(classDouble, midDoubleInit, (double)sum);
    jobject objAve = env->NewObject(classDouble, midDoubleInit, average);
    // Set to the jobjectArray
    env->SetObjectArrayElement(outJNIArray, 0, objSum);
    env->SetObjectArrayElement(outJNIArray, 1, objAve);

    return outJNIArray;
}



// Global Reference to the Java class "java.lang.Integer"
static jclass classInteger;
static jmethodID midIntegerInit;

jobject getInteger(JNIEnv *env, jobject thisObj, jint number)
{
   // Get a class reference for java.lang.Integer if missing
   if (NULL == classInteger)
   {
      cout << "Find java.lang.Integer" << endl;
      // FindClass returns a local reference
      jclass classIntegerLocal = env->FindClass("java/lang/Integer");
      // Create a global reference from the local reference
      classInteger = (jclass)env->NewGlobalRef(classIntegerLocal);
      // No longer need the local reference, free it!
      env->DeleteLocalRef(classIntegerLocal);
   }
   if (NULL == classInteger) return NULL;

   // Get the Method ID of the Integer's constructor if missing
   if (NULL == midIntegerInit)
   {
	  cout << "Get Method ID for java.lang.Integer's constructor" << endl;
      midIntegerInit = env->GetMethodID(classInteger, "<init>", "(I)V");
   }
   if (NULL == midIntegerInit) return NULL;

   // Call back constructor to allocate a new instance, with an int argument
   jobject newObj = env->NewObject(classInteger, midIntegerInit, number);
   cout << "In C++, constructed java.lang.Integer with number: " << number << endl;

   return newObj;
}

JNIEXPORT jobject JNICALL Java_HelloQT2JNI_getFirstIntegerObject(JNIEnv *env, jobject thisObj, jint number)
{
	return getInteger(env, thisObj, number);
}
JNIEXPORT jobject JNICALL Java_HelloQT2JNI_anotherGetIntegerObject(JNIEnv *env, jobject thisObj, jint number)
{
	return getInteger(env, thisObj, number);
}

