
public class HelloQT2JNI {

	static 
	{
		// Solaris: libHello.so
		// Linux: libHello.so
		// Windows: Hello.dll
		// Mac: libHello.jnilib
		System.loadLibrary("Hello");
	}
	
	// Quelle: https://www3.ntu.edu.sg/home/ehchua/programming/java/JavaNativeInterface.html
	
	// Declare native method
	private native void sayHello();
	
	// Native method that receives a Java String and return a Java String
	private native String sayHello(String msg);
	
	// Declare a native method average() that receives two ints and return a double containing the average
	private native double average(int n1, int n2);
	
	// Declare a native method sumAndAverage() that receives an int[] and
	// return a double[2] array with [0] as sum and [1] as average
	private native double[] sumAndAverage(int[] numbers);
	
	// Accessing Object's Instance Variables
	// Instance variables
	private int intNumber = 88;
	private String message = "Hello from Java";
	// Declare a native method that modifies the instance variables
	private native void modifyInstanceVariable();
	
	// Accessing Class' Static Variables
	// Static variables
	private static double doubleNumber = 55.66;
	// Declare a native method that modifies the static variable
	private native void modifyStaticVariable();
	
	// Callback Instance Methods and Static Methods
	// Declare a native method that calls back the Java methods below
	private native void nativeMethod();
	// To be called back by the native code
	private void callback() {
	   System.out.println("In Java");
	} 
	private void callback(String message) {
	   System.out.println("In Java with " + message);
	} 
	private double callbackAverage(int n1, int n2) {
	   return ((double)n1 + n2) / 2.0;
	} 
	// Static method to be called back
	private static String callbackStatic() {
	   return "From static Java method";
	}
	
	// Callback the Constructor to Create a New Java Object in the Native Code
	// Native method that calls back the constructor and return the constructed object.
	// Return an Integer object with the given int.
	private native Integer getIntegerObject(int number);
	
	// Array of Objects
	// Native method that receives an Integer[] and
	// returns a Double[2] with [0] as sum and [1] as average
	private native Double[] objSumAndAverage(Integer[] numbers);
	
	// Local and Global References
	// A native method that returns a java.lang.Integer with the given int.
	private native Integer getFirstIntegerObject(int number); 
	// Another native method that also returns a java.lang.Integer with the given int.
	private native Integer anotherGetIntegerObject(int number);
	   
	public static void main(String[] args) {
	
		System.out.println("> private native void sayHello():");
		new HelloQT2JNI().sayHello();  // invoke the native method
		System.out.println();
		
		System.out.println("> private native String sayHello(String msg):");
		String result = new HelloQT2JNI().sayHello("Hello from Java");
	    System.out.println("In Java, the returned string is: " + result);
	    System.out.println();
	    
	    System.out.println("> private native double average(int n1, int n2):");
	    System.out.println("In Java, the average is " + new HelloQT2JNI().average(3, 2));
	    System.out.println();
	    
	    System.out.println("> private native double[] sumAndAverage(int[] numbers):");
	    int[] numbers = {22, 33, 33};
	    double[] results = new HelloQT2JNI().sumAndAverage(numbers);
	    System.out.println("In Java, the sum is " + results[0]);
	    System.out.println("In Java, the average is " + results[1]);
	    System.out.println();
	    
	    System.out.println("> private native void modifyInstanceVariable():");
	    HelloQT2JNI instanceHelloJNI = new HelloQT2JNI();
	    instanceHelloJNI.modifyInstanceVariable();
	    System.out.println("In Java, int is " + instanceHelloJNI.intNumber); // 88
	    System.out.println("In Java, String is " + instanceHelloJNI.message); // Hello from Java
	    System.out.println();
	    
	    System.out.println("> private native void modifyStaticVariable():");
	    HelloQT2JNI staticHelloJNI = new HelloQT2JNI();
	    staticHelloJNI.modifyStaticVariable();
	    System.out.println("In Java, the double is " + doubleNumber);
	    System.out.println();
	    
	    System.out.println("> private native void nativeMethod():");
	    new HelloQT2JNI().nativeMethod();
	    System.out.println();
	    
	    System.out.println("> private native Integer getIntegerObject(int number):");
	    HelloQT2JNI obj = new HelloQT2JNI();
	    System.out.println("In Java, the number is " + obj.getIntegerObject(9999));
	    System.out.println();
	    
	    System.out.println("> private native Double[] objSumAndAverage(Integer[] numbers):");
	    Integer[] objNumbers = {11, 22, 32};  // auto-box
	    Double[] objResults = new HelloQT2JNI().objSumAndAverage(objNumbers);
	    System.out.println("In Java, the sum is " + objResults[0]);  // auto-unbox
	    System.out.println("In Java, the average is " + objResults[1]);
	    System.out.println();
	    
	    System.out.println("> private native Integer getFirstIntegerObject(int number):");
	    System.out.println("> private native Integer anotherGetIntegerObject(int number):");
	    HelloQT2JNI localGlobalHelloJNI = new HelloQT2JNI();
	    System.out.println(localGlobalHelloJNI.getFirstIntegerObject(1));
	    System.out.println(localGlobalHelloJNI.getFirstIntegerObject(2));
	    System.out.println(localGlobalHelloJNI.anotherGetIntegerObject(11));
	    System.out.println(localGlobalHelloJNI.anotherGetIntegerObject(12));
	    System.out.println(localGlobalHelloJNI.getFirstIntegerObject(3));
	    System.out.println(localGlobalHelloJNI.anotherGetIntegerObject(13));
	    System.out.println();
	}
}

